/* 
    Webcam Lab

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    (C) 2005 Jason Hunt
    nulluser@gmail.com
   
    file: canny.cpp

    This preforms canny edge detection using this method:
        
        
    1)  Apply a gauss smooting filter

    2) Calculate the Edge magnitude using a sobel filters in the
      x and y directions.

    3) Calculate the ratio of the y to x directional magnitudes.
       This is so the edge direction can be calculated without the 
       arctan function and by only apply two directional filters

    4) Threshold the gradient magnitude using P1

    5) Preform non-maximal sppression using the thresholded graident
        magnutude and the gradient ratio

*/


#include "core.h"





unsigned int *grad_mag = NULL;
int *grad_ratio = NULL;





#define LIM_LOW     0.4142
#define LIM_HIGH    2.1943



/* Get memory for canny edge detection system */
void canny_start( void )
{
    if (grad_mag != NULL) free(grad_mag);    

    grad_mag = (unsigned int *) malloc(size_x * size_y * sizeof(unsigned int));
        
    if (grad_mag  == NULL)
    {            
        add_line("Unable to create grad_mag buffer");
        return;
    }     

    if (grad_ratio != NULL) free(grad_ratio);    

    grad_ratio = (int *) malloc(size_x * size_y * sizeof(int));
        
    if (grad_ratio  == NULL)
    {            
        add_line("Unable to create grad_ratio buffer");
        return;
    }  
    
}
/* End of canny start */


/* Free canny memory */
void canny_stop (void )
{
    if (grad_mag) free(grad_mag);
    if (grad_ratio) free(grad_ratio);
}
/* End of canny_stop */


/* Use Sobel operator to calculate gridents */
void calulate_gradients(color_type *s)
{
    unsigned int index = size_x; // skip first line

    for (int y = 1; y < size_y-1; y++)
    {
        index++;
        for (int x = 1; x < size_x-1; x++)
        {
            // Point to the line above and one to the left
            unsigned int t = index - size_x - 1;
                   
            double x_sum = 0;
            double y_sum = 0;          
          
            unsigned int f = 0;

            // Preform convolution
            for (int i = 0; i < 3; i++)
            {
                for (int j = 0; j < 3; j++)
                {                          
                    double size = color_size(s[t]);
                    x_sum += size * sobel_x[f];
                    y_sum += size * sobel_y[f];
                                         
                    t++;  f++;                      
                }
                
                t += size_x - 3; // next line down, and 3 to left
            }

            grad_mag[index] = (unsigned int)(fabs(x_sum) + fabs(y_sum));

            // compute gradient  ratio.  Discard Y sign
            
            if (x_sum != 0)             // convert to fixed point
                grad_ratio[index] = (int) ( 65536.0 * fabs(y_sum) / x_sum);
            else
                grad_ratio[index] = 0;
               
            index++; 
        }
        index++; // Skip over unused right edge
    }    
}
/* End of calulate_gradients */


/* Preform non-maximal suppersion On a pixel */
unsigned char non_maximal(unsigned int index, unsigned int c)
{    
    if (c == 0) return(0);
    
    double m = grad_ratio[index] / 65536.0;       // Get the ratop
                
    // Here we are using the ratio of grad_mag.y / grad_mag.x
    // To determine the direction of the graident

    if (m > LIM_HIGH || m < -LIM_HIGH)      // up or down graident
    {
        if (grad_mag[index] > grad_mag[index - size_x] &&
            grad_mag[index] > grad_mag[index + size_x]) return(c);           
    } else
    if (m > LIM_LOW)                        // right diagnal graident             
    {                                       
        if (grad_mag[index] > grad_mag[index - size_x + 1] &&
            grad_mag[index] > grad_mag[index + size_x - 1]) return(c);
    } 
    if (m < -LIM_LOW)                       // left diagnal graident             
    {                             
        if (grad_mag[index] > grad_mag[index - size_x - 1] &&
            grad_mag[index] > grad_mag[index + size_x + 1]) return(c);  
   }
   
   // left / right graident
   
   if (grad_mag[index] > grad_mag[index - 1] &&
       grad_mag[index] > grad_mag[index + 1]) return(c);     
   
    return(0);
}
/* End of non_maximal */


/* Canny edge detection */
void canny_edge(color_type *s, color_type *d, double p1, double p2, double p3)
{
    transform_frame( s, d, gauss_blur, 5, 1/159.0, 0);

    memcpy(s, d, DATA_SIZE);

    calulate_gradients(s);

    unsigned int index = 0;

    // Skip unused line
    for (int x = 0; x < size_x; x++)
        d[index++] = black;

    for (int y = 1; y < size_y - 1; y++)    
    {
        // Skip unused pixel
        d[index] = black;
        index++;

        for (int x = 1; x < size_x - 1; x++)
        {
            unsigned int c = 0;
            unsigned int ranged_color = 0;

            // apply threshold
            if (grad_mag[index] > p1)
                c = (int)(grad_mag[index] + 128);

            c = non_maximal(index, c);

            // clamp
            if (c > 255) c = 255;
            if (c < 0) c = 0;

            // If the edge passed the threshold, set the pixel red
            // Otherwise set it to the value form the starting frame
            if (c == 0)
                d[index] = save_frame[index];
            else
                d[index] = red;
                    
            index++;
        }

        // Skip unused pixel;
        d[index] = black;
        index++;                
    }
    
    // Skip unused line
    for (int x = 0; x < size_x; x++)
        d[index++] = black;
}
/* End of canny_edge */

